/*
 *    TTTTTTTTTTTTTT  HH          HH  OOOOOOOOOOOOOO  MM          MM
 *    TTTTTTTTTTTTTT  HH          HH  OOOOOOOOOOOOOO  MMM        MMM
 *          TT        HH          HH  OO          OO  MMMM      MMMM 
 *          TT        HH          HH  OO          OO  MM MM    MM MM
 *          TT        HH          HH  OO          OO  MM  MM  MM  MM 
 *          TT        HHHHHHHHHHHHHH  OO          OO  MM   MMMM   MM
 *          TT        HHHHHHHHHHHHHH  OO          OO  MM    MM    MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OO          OO  MM          MM
 *          TT        HH          HH  OOOOOOOOOOOOOO  MM          MM
 *          TT        HH          HH  OOOOOOOOOOOOOO  MM          MM 
 *
 *                      l'mulateur Thomson TO7-70
 *
 *  Copyright (C) 1996 Sylvain Huet, 1999-2003 Eric Botcazou.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *  Module     : linux/disk.c
 *  Version    : 1.5.5
 *  Cr par   : Eric Botcazou 29/07/2000
 *  Modifi par: Eric Botcazou 05/11/2003
 *
 *  Lecture directe des disquettes Thomson.
 */


#ifndef SCAN_DEPEND
   #include <stdio.h>
   #include <fcntl.h>
   #include <unistd.h>
   #include <sys/ioctl.h>
   #include <linux/fd.h>
   #include <linux/fdreg.h>
#endif

#include "to7.h"


#define DISK_RETRY   3

static int fd[2] = {-1, -1};
static int drive_type[2];

#define IS_5_INCHES(drive) ((drive_type[drive]>0) && (drive_type[drive]<3))



/* ResetFloppyDrive:
 *  Rinitialise le lecteur de disquettes.
 */
static void ResetFloppyDrive(int drive)
{
    struct floppy_struct fd_prm;

    fd_prm.head    = 2;
    fd_prm.track   = (IS_5_INCHES(drive) ? 40 : 80);
    fd_prm.sect    = 8;
    fd_prm.size    = fd_prm.head * fd_prm.track * fd_prm.sect;
    fd_prm.stretch = 0;
    fd_prm.gap     = 0x1B;
    fd_prm.rate    = 0x3A;
    fd_prm.spec1   = 0xDF;
    fd_prm.fmt_gap = 0x2C;

    ioctl(fd[drive], FDSETPRM, &fd_prm);
}



/* OpenFloppyDrive:
 *  Obtient le descripteur de fichier pour le lecteur de disquettes.
 */
static int OpenFloppyDrive(int drive)
{
    char dev_str[16];

    snprintf(dev_str, sizeof(dev_str), "/dev/fd%d", drive);

    if (to7_DirectWriteSector)
    {
        if ((fd[drive]=open(dev_str, O_RDWR | O_NDELAY))<0)
        {
            to7_DirectWriteSector = NULL;
            to7_DirectFormatTrack = NULL;
        }
        else
            goto success;
    }

    if ((fd[drive]=open(dev_str, O_RDONLY | O_NDELAY))<0)
        to7_DirectReadSector = NULL;
    else
        goto success;

    return 0;

  success:
     ResetFloppyDrive(drive);
     return 1;
}



/* ExecCommand:
 *  Excute la commande spcifie via l'appel ioctl() FDRAWCMD.
 */
static int ExecCommand(int drive, struct floppy_raw_cmd *fd_cmd)
{
    int i, ret;

    if (fd[drive]<0 && !OpenFloppyDrive(drive))
        return 0x10;  /* lecteur non prt */

    for (i=0; i<DISK_RETRY; i++)
    {
        ret=ioctl(fd[drive], FDRAWCMD, fd_cmd);

        if (ret >= 0)
            break;

        ResetFloppyDrive(drive);
    }

    if (ret<0)
        return 0x10;  /* lecteur non prt */

#ifdef DEBUG
    printf("fd_cmd reply: ");
    for (i=0; i<3; i++)
        printf("ST%d=%02x ", i, fd_cmd->reply[i]);
    printf("\n");
#endif

    switch (fd_cmd->reply[1])  /* ST1 */
    {
        case 0x01:  /* Missing Address Mark */
            return 0x04;   /* erreur sur l'adresse */

        case 0x02:  /* Write Protected */
            return 0x01;   /* disk protg en criture */

        case 0x04:  /* No Data - unreadable */
            return 0x08;   /* erreur sur les donnes */

        case 0x20:  /* CRC error in data or addr */
            if (fd_cmd->reply[2]==0x20)
                return 0x08;   /* erreur sur les donnes */
            else
                return 0x04;   /* erreur sur l'adresse */

        default:
            return 0;  /* OK */
    }
}



/* ReadSector:
 *  Lit le secteur spcifi sur la disquette.
 */
static int ReadSector(int drive, int track, int sector, int nsects, unsigned char data[])
{
    struct floppy_raw_cmd fd_cmd;
    int pc_drive = drive/2;

    /* paramtres de commande */
    fd_cmd.flags  = FD_RAW_READ | FD_RAW_INTR | FD_RAW_NEED_SEEK;
    fd_cmd.data   = data;
    fd_cmd.length = 256*nsects; /* SECTOR_SIZE */
    fd_cmd.rate   = IS_5_INCHES(pc_drive) ? 1 : 2;
    fd_cmd.track  = IS_5_INCHES(pc_drive) ? track*2 : track;

    fd_cmd.cmd[0] = FD_READ;
    fd_cmd.cmd[1] = (drive%2) << 2;
    fd_cmd.cmd[2] = track;
    fd_cmd.cmd[3] = 0;  /* head */
    fd_cmd.cmd[4] = sector;
    fd_cmd.cmd[5] = 1;  /* SECTOR_SIZE >> 8 */
    fd_cmd.cmd[6] = 16; /* num sectors      */
    fd_cmd.cmd[7] = 0x1B;
    fd_cmd.cmd[8] = 0xFF;
    fd_cmd.cmd_count = 9;

    return ExecCommand(pc_drive, &fd_cmd);
}



/* WriteSector:
 *  Ecrit le secteur spcifi sur la disquette.
 */
static int WriteSector(int drive, int track, int sector, int nsects, const unsigned char data[])
{
    struct floppy_raw_cmd fd_cmd;
    int pc_drive = drive/2;

    /* paramtres de commande */
    fd_cmd.flags  = FD_RAW_WRITE | FD_RAW_INTR | FD_RAW_NEED_SEEK;
    fd_cmd.data   = (unsigned char *)data;
    fd_cmd.length = 256*nsects; /* SECTOR_SIZE */
    fd_cmd.rate   = IS_5_INCHES(pc_drive) ? 1 : 2;
    fd_cmd.track  = IS_5_INCHES(pc_drive) ? track*2 : track;

    fd_cmd.cmd[0] = FD_WRITE;
    fd_cmd.cmd[1] = (drive%2) << 2;
    fd_cmd.cmd[2] = track;
    fd_cmd.cmd[3] = 0;  /* head */
    fd_cmd.cmd[4] = sector;
    fd_cmd.cmd[5] = 1;  /* SECTOR_SIZE >> 8 */
    fd_cmd.cmd[6] = 16; /* num sectors      */
    fd_cmd.cmd[7] = 0x1B;
    fd_cmd.cmd[8] = 0xFF;
    fd_cmd.cmd_count = 9;

    return ExecCommand(pc_drive, &fd_cmd);
}



/* FormatTrack:
 *  Formate la piste en utilisant la table des headers spcifie.
 */
static int FormatTrack(int drive, int track, const unsigned char header_table[])
{
    struct floppy_raw_cmd fd_cmd;
    int pc_drive = drive/2;

    /* paramtres de commande */
    fd_cmd.flags  = FD_RAW_WRITE | FD_RAW_INTR | FD_RAW_NEED_SEEK;
    fd_cmd.data   = (unsigned char *)header_table;
    fd_cmd.length = 64;
    fd_cmd.rate   = IS_5_INCHES(pc_drive) ? 1 : 2;
    fd_cmd.track  = IS_5_INCHES(pc_drive) ? track*2 : track;

    fd_cmd.cmd[0] = FD_FORMAT;
    fd_cmd.cmd[1] = (drive%2) << 2;
    fd_cmd.cmd[2] = 1;     /* SECTOR_SIZE >> 8 */
    fd_cmd.cmd[3] = 16;    /* num sectors      */
    fd_cmd.cmd[4] = 0x2C;  /* fmt_gap          */
    fd_cmd.cmd[5] = 0xE5;  /* filler_byte      */
    fd_cmd.cmd_count = 6;

    return ExecCommand(pc_drive, &fd_cmd);
}



/* InitDirectDisk:
 *  Initialise le module de lecture directe.
 */
int InitDirectDisk(int to_drive_type[4], int enable_write)
{
    struct floppy_drive_params fd_params;
    char dev_str[16];
    int i, num_drives = 0;

    for (i=0; i<2; i++)
    {
        /* get drive type */
        snprintf(dev_str, sizeof(dev_str), "/dev/fd%d", i);

        if ((fd[i]=open(dev_str, O_RDONLY | O_NDELAY))<0)
        {
            drive_type[i] = 0;

            to_drive_type[2*i] = 0;
            to_drive_type[2*i+1] = 0;
        }
        else
        {
            ioctl(fd[i], FDGETDRVPRM, &fd_params);
            close(fd[i]);
            fd[i] = -1;

            /* drive type: 1 (5"25 - 360 kb)
             *             2 (5"25 - 1.2 Mb)
             *             3 (3"5  - 720 kb)
             *             4 (3"5 - 1.44 Mb)
             *             5 (3"5 - 2.88 Mb)
             *             6 (3"5 - 2.88 Mb)
             */
            if (fd_params.cmos > 6)
            {
                drive_type[i] = 0;

                to_drive_type[2*i] = 0;
                to_drive_type[2*i+1] = 0;
            }
            else
            {
                drive_type[i] = fd_params.cmos;

                to_drive_type[2*i] = fd_params.cmos;
                to_drive_type[2*i+1] = fd_params.cmos;
            }

            num_drives++;
        }
    }

    to7_DirectReadSector = ReadSector;

    if (enable_write)
    {
        to7_DirectWriteSector = WriteSector;
        to7_DirectFormatTrack = FormatTrack;
    }

    return num_drives;
}



/* ExitDirectDisk:
 *  Met au repos le module de lecture directe.
 */
void ExitDirectDisk(void)
{
# if 0 //LUDO:
    int i;

    to7_DirectReadSector = NULL;
    to7_DirectWriteSector = NULL;
    to7_DirectFormatTrack = NULL;

    for (i=0; i<2; i++)
    {
        if (fd[i] >= 0)
        {
            close(fd[i]);
            fd[i] = -1;
        }
    }
# endif
}

